from pathlib import Path, PosixPath
import numpy as np

from .log import logger
from .recording import Recording
from .Event import Event
from breesy.load_events import find_and_load_event_for_file
from .RecordingMetadata import RecordingMetadata
from .errors import BreesyError, BreesyInternalError, protect_from_lib_error
from .type_hints import enforce_type_hints

# Note: if changing this, then update load_dataset if/else statement as well
SUPPORTED_FILE_EXTENSIONS = ['.csv', '.mat', '.tsv', '.npy', '.eeg', '.vhdr', '.avg', '.seg', '.fdt', '.set', '.edf', '.xdf', '.bdf']


@enforce_type_hints
def load_recording(filename: str | Path | PosixPath, sample_rate: int | float | None = None) -> Recording:
    """Load a recording from a file with a supported extension.

    :param filename: Path to the file to load
    :param sample_rate: Optional sampling rate in Hz

    :return: Recording object containing the loaded data
    """
    filename = str(filename)
    file_extension = Path(filename).suffix
    if not file_extension:
        raise BreesyError("File name should contain an extension",
                          "Provide a file with a supported extension, example: 'my_data.gdf'")

    is_supported_extension = file_extension in SUPPORTED_FILE_EXTENSIONS
    if not is_supported_extension:
        raise BreesyError(f"Unsupported file extension: {file_extension}",
                          "Provide a file with a supported extension, example: 'my_data.gdf'. "
                          f"Currently supported extensions: {SUPPORTED_FILE_EXTENSIONS}")

    path_filename = Path(filename)
    is_valid_filename = Path.is_file(path_filename)
    if not is_valid_filename:
        raise BreesyError(
            f"File not found: {filename}", "a) Make sure that the file is in your working directory; "
                                           f"the current working directory is: {Path.cwd()}. "
                                           f"b) Double check that your file explorer is showing the full file name "
                                           f"- Windows does not show file extensions by default.")

    if file_extension == '.xdf':
        from .load_xdf import load_xdf
        return load_xdf(filename)
    elif file_extension == '.edf':
        from .load_edf import load_edf
        return load_edf(filename)
    elif file_extension == '.bdf':
        from .load_edf import load_bdf
        return load_bdf(filename)
    elif file_extension in ['.eeg', '.avg', '.seg', '.vhdr']:
        from .load_brainvision import load_brainvision_eeg
        if sample_rate is not None:
            logger.warning(f"Ignoring provided sample_rate of {sample_rate} because BrainVision files contain their own sample rate information.")
        return load_brainvision_eeg(filename)
    elif file_extension == '.csv':
        return load_csv(filename, sample_rate=sample_rate)
    elif file_extension == '.mat':
        from .load_matlab import load_mat
        return load_mat(filename, sample_rate=sample_rate)
    elif file_extension == '.tsv':
        return load_tsv(filename, sample_rate=sample_rate)
    elif file_extension == '.npy':
        return load_numpy(filename, sample_rate=sample_rate)
    elif file_extension in ['.set', '.fdt']:
        from .load_eeglab import load_eeglab
        if sample_rate is not None:
            logger.warning(f"Ignoring provided sample_rate of {sample_rate} because EEGLab files contain their own sample rate information.")
        return load_eeglab(filename)
    else:
        raise BreesyInternalError("File extension valid but not handled")


@enforce_type_hints
def load_delimited_text(filename: str, delimiter: str = ',', sample_rate: int | float | None = None) -> Recording:
    """Load a delimited text file (CSV, TSV, etc.).

    :param filename: Path to the text file
    :param delimiter: Character used to separate values in the file
    :param sample_rate: Sampling rate in Hz. If not provided, will attempt to guess from time column. If sample rate cannot be guessed, an error will be raised.

    :return: Recording object containing the loaded data
    """
    import csv

    # First pass: read headers and determine structure
    with open(filename, 'r', newline='') as file:
        reader = csv.reader(file, delimiter=delimiter)
        headers = next(reader)  # Get column names

        # Check if first column might be time
        time_column_idx = None
        if headers[0].lower() in ['time', 't', 'timestamp']:
            time_column_idx = 0

        # Exclude time column from channel names if present
        if time_column_idx is not None:
            channel_names = headers[1:]
        else:
            channel_names = headers

        # Count rows for pre-allocation
        row_count = sum(1 for _ in reader)

    # Second pass: read all data
    with open(filename, 'r', newline='') as file:
        reader = csv.reader(file, delimiter=delimiter)
        next(reader)  # Skip header

        # Pre-allocate data array
        data = np.zeros((len(channel_names), row_count))
        time_values = None

        if time_column_idx is not None:
            time_values = np.zeros(row_count)

        # Fill data
        for i, row in enumerate(reader):
            if i >= row_count:
                break

            # Convert values to float, handling potential errors
            try:
                row_values = [float(val) for val in row]
            except ValueError:
                # Skip rows that can't be converted to float
                continue

            if time_column_idx is not None:
                time_values[i] = row_values[time_column_idx]
                data_values = row_values[:time_column_idx] + row_values[time_column_idx + 1:]
            else:
                data_values = row_values

            for j, val in enumerate(data_values):
                if j < len(channel_names):
                    data[j, i] = val

    if sample_rate is not None:
        sample_rate = float(sample_rate)
    else:
        if time_values is not None and len(time_values) > 1:
            # Estimate sample rate
            time_diffs = np.diff(time_values)
            avg_diff = float(np.mean(time_diffs))
            if avg_diff > 0:
                sample_rate = float(round(1.0 / avg_diff))
        
        if sample_rate is None:
            raise BreesyError(
                "Sample rate was not provided and it could not be guessed from the data. "
                "Please provide the sample rate manually by passing the 'sample_rate' parameter.",
                "Provide the sample rate as a named argument to the function, e.g.: load_delimited_text('file.csv', sample_rate=500)"
            )


    # Look for event columns (columns with mostly 0s and few 1s)
    events = []
    for i, col_name in enumerate(channel_names):
        col_data = data[i, :]
        unique_vals = set(np.unique(col_data))
        # Check if column is binary and sparse (potential event marker)
        if unique_vals <= {0, 1} and np.mean(col_data) < 0.1:
            event_indices = np.where(col_data == 1)[0]
            for idx in event_indices:
                events.append(Event(name=col_name, index=idx))

    data, channel_names = _leave_only_eeg_channels(data, channel_names)

    metadata = RecordingMetadata(
        name=Path(filename).stem,
        file_path=Path(filename),
        description=f"Data loaded from {filename}"
    )

    return Recording(
        data=data,
        channel_names=channel_names,
        sample_rate=sample_rate,
        events=events,
        metadata=metadata
    )

# wrapper function for backward compatibility
@enforce_type_hints
def load_csv(filename: str) -> Recording:
    """Load a CSV file.

    :param filename: Path to the CSV file

    :return: Recording object containing the loaded data
    """
    return load_delimited_text(filename, delimiter=',')


# wrapper function for backward compatibility
@enforce_type_hints
def load_tsv(filename: str) -> Recording:
    """Load a TSV file.

    :param filename: Path to the TSV file

    :return: Recording object containing the loaded data
    """
    return load_delimited_text(filename, delimiter='\t')


@enforce_type_hints
def load_numpy(filename: str, sample_rate: int | float | None = None, channel_names: list[str] | None = None) -> Recording:
    """Load a numpy .npy file.

    :param filename: Path to the .npy file
    :param sample_rate: Sampling rate in Hz. If not provided, will attempt to detect from time column.
                        If no time column is found and sample_rate is None, an error will be raised.
    :param channel_names: List of channel names. If not provided, default names will be generated.

    :return: Recording object containing the loaded data
    """
    data = _np_load(filename)

    # TODO: #12 we need to handle this case properly, currently just taking first trial
    if data.ndim == 3:
        # likely many recordings instead of one
        logger.warning("The loaded NumPy array has 3 dimensions, which likely means it contains multiple recordings; loading only the first one")
        data = data[0]

    if data.ndim != 2:
        raise BreesyError(
            f"The NumPy array has {data.ndim} dimensions, but 2 dimensions are required.",
            "Make sure the NumPy file contains a 2D array with shape (channels, samples)."
        )

    # if the first dimension is larger, assume it's (samples, channels) and transpose
    if data.shape[0] > data.shape[1]:
        logger.warning("The first data dimension seems to be the samples, because data contains more rows than columns. Dimensions will be switched places.")
        data = data.T

    extracted_sample_rate = None
    time_column_idx = _detect_time_column(data)
    if time_column_idx is not None:
        extracted_sample_rate = _calculate_sample_rate_from_time_column(data[time_column_idx])
        logger.info(f"Detected time column at index {time_column_idx}. Extracted sample rate: {extracted_sample_rate} Hz")
        # remove time column from data
        data = np.delete(data, time_column_idx, axis=0)

    # Determine final sample rate
    if sample_rate is not None:
        if extracted_sample_rate is not None and abs(sample_rate - extracted_sample_rate) > 0.01:
            logger.warning(f"Warning: Provided sample_rate ({sample_rate} Hz) differs from extracted sample rate ({extracted_sample_rate} Hz). Using the provided sample_rate.")
        sample_rate = float(sample_rate)
    elif extracted_sample_rate is not None:
        sample_rate = extracted_sample_rate
    else:
        raise BreesyError(
            "No sample rate provided and could not detect sample rate from data.",
            "Please provide the sample rate manually by passing the 'sample_rate' parameter. "
            "Example: load_numpy(filename, sample_rate=250)"
        )

    # Create channel names
    if channel_names is None:
        channel_names = [f'Channel_{i+1}' for i in range(data.shape[0])]
    elif len(channel_names) != data.shape[0]:
        raise BreesyError(
            f"The number of channel names ({len(channel_names)}) does not match the number of channels ({data.shape[0]}) in the data.",
            "Make sure that the number of channel names matches the number of channels in the data. "
            f"Expected {data.shape[0]} channel names."
        )

    events = find_and_load_event_for_file(filename, sample_rate)

    data, channel_names = _leave_only_eeg_channels(data, channel_names)

    metadata = RecordingMetadata(
        name=Path(filename).stem,
        file_path=Path(filename),
        description=f"NumPy data loaded from {filename}"
    )

    return Recording(
        data=data,
        channel_names=channel_names,
        sample_rate=sample_rate,
        events=events,
        metadata=metadata
    )


@enforce_type_hints
def from_array(data: np.ndarray, sample_rate: int | float | None = None, channel_names: list[str] | None = None) -> Recording:
    """Convert a numpy array into a Recording object.

    :param data: NumPy array containing EEG data. Should be 2D with shape (channels, samples) or (samples, channels).
                 If a time column is detected, it will be used to extract the sample rate and removed from the data.
    :param sample_rate: Sampling rate in Hz. If not provided, will attempt to detect from time column.
                        If no time column is found and sample_rate is None, an error will be raised.
    :param channel_names: List of channel names. If not provided, default names will be generated.

    :return: Recording object containing the loaded data
    """
    # Ensure it's a 2D array
    if data.ndim != 2:
        raise BreesyError(
            f"The NumPy array has {data.ndim} dimensions, but 2 dimensions are required.",
            "Make sure the NumPy array is 2D with shape (channels, samples) or (samples, channels)."
        )

    # Auto-transpose if the first dimension is larger (likely samples, channels)
    if data.shape[0] > data.shape[1]:
        logger.warning("The first data dimension seems to be the samples, because data contains more rows than columns. Dimensions will be switched places.")
        data = data.T

    extracted_sample_rate = None
    time_column_idx = _detect_time_column(data)
    if time_column_idx is not None:
        extracted_sample_rate = _calculate_sample_rate_from_time_column(data[time_column_idx])
        logger.info(f"Detected time column at index {time_column_idx}. Extracted sample rate: {extracted_sample_rate} Hz")
        # remove time column from data
        data = np.delete(data, time_column_idx, axis=0)

    # Determine final sample rate
    if sample_rate is not None:
        if extracted_sample_rate is not None and abs(sample_rate - extracted_sample_rate) > 0.01:
            logger.warning(f"Warning: Provided sample_rate ({sample_rate} Hz) differs from extracted sample rate ({extracted_sample_rate} Hz). Using the provided sample_rate.")
        sample_rate = float(sample_rate)
    elif extracted_sample_rate is not None:
        sample_rate = extracted_sample_rate
    else:
        raise BreesyError(
            "No sample rate provided and could not detect sample rate from data.",
            "Please provide the sample rate manually by passing the 'sample_rate' parameter. "
            "Example: from_array(data, sample_rate=250)"
        )

    # Create channel names
    if channel_names is None:
        channel_names = [f'Channel_{i+1}' for i in range(data.shape[0])]
    elif len(channel_names) != data.shape[0]:
        raise BreesyError(
            f"The number of channel names ({len(channel_names)}) does not match the number of channels ({data.shape[0]}) in the data.",
            "Make sure that the number of channel names matches the number of channels in the data. "
            f"Expected {data.shape[0]} channel names."
        )
    
    data, channel_names = _leave_only_eeg_channels(data, channel_names)

    # Create metadata
    metadata = RecordingMetadata(
        name="NumPy Recording",
        file_path=None,
        description="Recording created from NumPy array"
    )

    return Recording(
        data=data,
        channel_names=channel_names,
        sample_rate=sample_rate,
        events=None,
        metadata=metadata
    )


# @enforce_type_hints
# def load_all_files_in_folder_by_extension(file_extension: str, folder_path: str = ""):
#     """Load all files with a specific extension from a folder.

#     :param file_extension: File extension to filter by (e.g., '.mat')
#     :param folder_path: Path to the folder containing the files

#     :return: List of Recording objects containing the loaded data
#     """
#     if not isinstance(folder_path, str):
#         raise BreesyError("Invalid folder_path argument type",
#                           "Provide a string as the folder path, example: 'data_folder', or '' (empty string or period) for current working directory.")

#     if not isinstance(file_extension, str):
#         raise BreesyError("Invalid file_extension argument type",
#                           "Provide a string as the file extension, example: '.gdf'")

#     folder = Path(folder_path)
#     if not folder.is_dir():
#         raise BreesyError(f"The folder path was not found: {folder.resolve()}",
#                           f"Check the folder path manually to make sure that the folder actually exists.")

#     full_path = folder.resolve()

#     files = list(full_path.glob(f"*{file_extension}"))
#     if not files:
#         raise BreesyError(
#             f"No {file_extension} files found in folder: '{folder_path}' (resolved absolute path: {folder.resolve()})",
#             f"Check the folder to make sure that it actually contains {file_extension} files.")

#     data_list = []
#     for file in files:
#         data = load_dataset(str(file))
#         data_list.append(data)

#     return data_list


# --------- Helper functions ---------

@protect_from_lib_error("numpy")
def _np_load(filename):
    return np.load(filename)


def _detect_time_column(data: np.ndarray) -> int | None:
    for i in range(data.shape[0]):
        if _is_likely_time_column(data[i]):
            return i
    return None


def _calculate_sample_rate_from_time_column(data: np.ndarray) -> float:
    if (data.size < 2) or data[0] == data[1]:
        raise BreesyInternalError(
            "Cannot calculate sample rate from time column with less than 2 unique values. Make sure the time column contains valid time values."
        )
    time_step = float(data[1] - data[0])
    sample_rate = float(1 / time_step)
    return sample_rate


def _is_likely_eeg_column(column: np.ndarray) -> bool:
    return (not _column_is_monotonic(column)) and (not _column_is_sparse(column))


def _is_likely_time_column(column: np.ndarray) -> bool:
    u_diffs = np.unique(np.diff(column).round(6))
    if len(u_diffs) == 1 and u_diffs[0] > 0:
        return True
    return False


def _leave_only_eeg_channels(data: np.ndarray, ch_names: list[str]) -> tuple[np.ndarray, list[str]]:
    eeg_ix = [i for i in range(data.shape[0]) if _is_likely_eeg_column(data[i])]
    removed = [name for i, name in enumerate(ch_names) if i not in eeg_ix]
    if removed:
        logger.warning(f"Ignored {len(removed)} channel(s) that don't look like EEG signal: {removed}")
    
    new_channel_names = [x for i, x in enumerate(ch_names) if i in eeg_ix]
    return data[eeg_ix], new_channel_names


def _column_is_monotonic(column: np.ndarray) -> bool:
    diffs = np.diff(column)
    return bool(np.all(diffs >= 0))


def _column_is_sparse(column: np.ndarray, sparsity_threshold: float = 0.2) -> bool:
    u_values, u_counts = np.unique(column, return_counts=True)
    most_common_value_rate = np.max(u_counts) / len(column)
    return bool(most_common_value_rate > sparsity_threshold)


def _find_keys_in_data(keys: list[str], searching_for: list[str]) -> list:
    found = []
    translation = str.maketrans('', '', '_- .')
    to_check = [k.lower().translate(translation) for k in searching_for]
    for k in keys:
        if k.lower().translate(translation) in to_check:
            found.append(k)
    return found



import numpy as np

from breesy import breesy_scipy, constants
from breesy.recording import Recording
from breesy.errors import BreesyError, BreesyInternalError
from breesy.recording import update_data
from breesy.type_hints import enforce_type_hints


BANDPASS_PARAMS = """
:param recording: Input recording
:param low: Lower frequency bound in Hz
:param high: Upper frequency bound in Hz
"""

SINGLE_FREQ_PARAMS = """
:param recording: Input recording
:param frequency: Frequency threshold in Hz
"""

NOTCH_PARAMS = """
:param recording: Input recording
:param frequency: Frequency to remove in Hz
"""

DEFAULT_WINDOW_PARAMS = constants.DEFAULT_WINDOW

_DOC_PARAMS = {
    name: value for name, value in globals().items()
    if name.endswith('_PARAMS') and isinstance(value, str)
}


def _format_doc(func):
    """Decorator that expands shared parameter docs in the docstring."""
    if func.__doc__:
        func.__doc__ = func.__doc__.format(**_DOC_PARAMS)
    return func


# Default filters


@_format_doc
@enforce_type_hints
def notch_filter(recording: Recording, frequency: int | float, smoothness_factor: int | float = constants.NOTCH_DEFAULT_SMOOTHNESS) -> Recording:
    """Apply notch filter to remove a specific frequency from a recording.
    By default, it is a FIR filter.

    {NOTCH_PARAMS}
    :param smoothness_factor: Controls the smoothness of attenuation. Higher value means wider affected (attenuated) band.

    :return: Filtered recording
    """
    _assert_frequency_exists(sample_rate=recording.sample_rate, freq=frequency)
    transition_width, notch_width = _get_fir_params_from_smoothness(smoothness_factor=smoothness_factor, notch_freq=frequency)

    new_data = recording.data.copy()
    new_data = _notch_fir(data=new_data, frequency=frequency, sample_rate=recording.sample_rate, transition_width=transition_width, notch_width=notch_width)
    return update_data(recording, new_data) 


@_format_doc
@enforce_type_hints
def bandpass_filter(recording: Recording, low: float | int, high: float | int,
                    smoothness_factor: int | float = constants.BANDPASS_DEFAULT_SMOOTHNESS) -> Recording:
    """Filter recording to keep only a specific frequency band by applying a bandpass filter.
    By default, it is a FIR filter.

    {BANDPASS_PARAMS}
    :param smoothness_factor: Controls the smoothness of attenuation. Higher value means wider transition region (from not attenuated to attenuated) from each side of the band.

    :return: Filtered recording
    """
    _assert_bandpass_low_high(sample_rate=recording.sample_rate, low=low, high=high)

    new_data = recording.data.copy()
    new_data = _bandpass_fir(data=new_data, low=low, high=high, sample_rate=recording.sample_rate,
                             transition_width=smoothness_factor)  # TODO: check if needs conversion
    return update_data(recording, new_data) 


@_format_doc
@enforce_type_hints
def lowpass_filter(recording: Recording, frequency: float | int,
                   smoothness_factor: int | float = constants.BANDPASS_DEFAULT_SMOOTHNESS) -> Recording:
    """Filter recording to keep only frequencies below certain threshold by applying a lowpass filter.
    By default, it is a FIR filter.

    {SINGLE_FREQ_PARAMS}
    :param smoothness_factor: Controls the smoothness of attenuation. Higher value means wider transition region (from not attenuated to attenuated) from each side of the band.

    :return: Filtered recording
    """
    _assert_frequency_exists(sample_rate=recording.sample_rate, freq=frequency)

    new_data = recording.data.copy()
    new_data = _bandpass_fir(data=new_data, low=None, high=frequency, sample_rate=recording.sample_rate,
                             transition_width=smoothness_factor)
    return update_data(recording, new_data) 


@_format_doc
@enforce_type_hints
def highpass_filter(recording: Recording, frequency: float | int,
                   smoothness_factor: int | float = constants.BANDPASS_DEFAULT_SMOOTHNESS) -> Recording:
    """Filter recording to keep only frequencies above certain threshold by applying a highpass filter.
    By default, it is a FIR filter.

    {SINGLE_FREQ_PARAMS}
    :param smoothness_factor: Controls the smoothness of attenuation. Higher value means wider transition region (from not attenuated to attenuated) from each side of the band.

    :return: Filtered recording
    """
    _assert_frequency_exists(sample_rate=recording.sample_rate, freq=frequency)

    new_data = recording.data.copy()
    new_data = _bandpass_fir(data=new_data, low=frequency, high=None, sample_rate=recording.sample_rate,
                             transition_width=smoothness_factor)
    return update_data(recording, new_data) 


# Specific filters


@_format_doc
@enforce_type_hints
def notch_filter_fir(recording: Recording, frequency: int | float, notch_width: int | float, transition_width: int | float,
                     window: str | None = None) -> Recording:
    """Apply FIR notch filter to remove a specific frequency from a recording.

    {NOTCH_PARAMS}
    :param notch_width: Controls the width of removed (strongly attenuated) band
    :param transition_width: Controls the width of transition region (from not attenuated to attenuated) from each side of removed band
    :param window: Name of window type to use (default: {DEFAULT_WINDOW_PARAMS})

    :return: Filtered recording
    """
    _assert_frequency_exists(sample_rate=recording.sample_rate, freq=frequency)
    new_data = recording.data.copy()
    new_data = _notch_fir(data=new_data, frequency=frequency, sample_rate=recording.sample_rate, 
                          transition_width=transition_width, notch_width=notch_width, window=window)
    return update_data(recording, new_data) 


@_format_doc
@enforce_type_hints
def notch_filter_iir(recording: Recording, frequency: int | float, quality_factor: int) -> Recording:
    """Apply IIR notch filter to remove a specific frequency from a recording.

    {NOTCH_PARAMS}
    :param quality_factor: Controls the strength and width of attenuation. Higher value means narrower and deeper removed band.

    :return: Filtered recording
    """
    _assert_frequency_exists(sample_rate=recording.sample_rate, freq=frequency)
    new_data = recording.data.copy()
    new_data = _notch_iir(data=new_data, frequency=frequency, sample_rate=recording.sample_rate, Q=quality_factor)
    return update_data(recording, new_data) 


@_format_doc
@enforce_type_hints
def bandpass_filter_fir(recording: Recording, low: float | int, high: float | int, transition_width: int | float,
                        window: str | None = None) -> Recording:
    """Apply FIR bandpass filter to keep only a specific frequency band.

    {BANDPASS_PARAMS}
    :param transition_width: Controls the width of transition region (from not attenuated to attenuated) from each side of the band
    :param window: Name of window type to use (default: {DEFAULT_WINDOW_PARAMS})
    
    :return: Filtered recording
    """
    _assert_bandpass_low_high(sample_rate=recording.sample_rate, low=low, high=high)
    new_data = recording.data.copy()
    new_data = _bandpass_fir(data=new_data, low=low, high=high, sample_rate=recording.sample_rate,
                             transition_width=transition_width, window=window)
    return update_data(recording, new_data) 


@_format_doc
@enforce_type_hints
def lowpass_filter_fir(recording: Recording, frequency: float | int, transition_width: int | float,
                       window: str | None = None) -> Recording:
    """Apply FIR bandpass filter to keep only frequencies below the threshold.

    {SINGLE_FREQ_PARAMS}
    :param transition_width: Controls the width of transition region (from not attenuated to attenuated)
    :param window: Name of window type to use (default: {DEFAULT_WINDOW_PARAMS})

    :return: Filtered recording
    """
    _assert_frequency_exists(sample_rate=recording.sample_rate, freq=frequency)
    new_data = recording.data.copy()
    new_data = _bandpass_fir(data=new_data, low=None, high=frequency, sample_rate=recording.sample_rate, transition_width=transition_width)
    return update_data(recording, new_data) 


@_format_doc
@enforce_type_hints
def highpass_filter_fir(recording: Recording, frequency: float | int, transition_width: int | float,
                        window: str | None = None) -> Recording:
    """Apply FIR bandpass filter to keep only frequencies above the threshold.

    {SINGLE_FREQ_PARAMS}
    :param transition_width: Controls the width of transition region (from attenuated to not attenuated)
    :param window: Name of window type to use (default: {DEFAULT_WINDOW_PARAMS})

    :return: Filtered recording
    """
    _assert_frequency_exists(sample_rate=recording.sample_rate, freq=frequency)
    new_data = recording.data.copy()
    new_data = _bandpass_fir(data=new_data, low=frequency, high=None, sample_rate=recording.sample_rate, transition_width=transition_width)
    return update_data(recording, new_data) 


@_format_doc
@enforce_type_hints
def bandpass_filter_iir(recording: Recording, low: float | int, high: float | int, filter_order: int) -> Recording:
    """Apply IIR bandpass filteer to keep only a specific frequency band.

    {BANDPASS_PARAMS}
    :param filter_order: Controls the strength of filtering. Higher values mean narrower transition regions from each side of the band.

    :return: Filtered recording
    """
    _assert_bandpass_low_high(sample_rate=recording.sample_rate, low=low, high=high)
    new_data = recording.data.copy()
    new_data = _bandpass_iir(data=new_data, low=low, high=high, sample_rate=recording.sample_rate, filter_order=filter_order)
    return update_data(recording, new_data) 


@_format_doc
@enforce_type_hints
def lowpass_filter_iir(recording: Recording, frequency: float | int, filter_order: int) -> Recording:
    """Apply IIR bandpass filteer to keep only frequencies below the threshold.

    {SINGLE_FREQ_PARAMS}
    :param filter_order: Controls the strength of filtering. Higher values mean narrower transition region.

    :return: Filtered recording
    """
    _assert_frequency_exists(sample_rate=recording.sample_rate, freq=frequency)
    new_data = recording.data.copy()
    new_data = _bandpass_iir(data=new_data, low=None, high=frequency, sample_rate=recording.sample_rate, filter_order=filter_order)
    return update_data(recording, new_data) 


@_format_doc
@enforce_type_hints
def highpass_filter_iir(recording: Recording, frequency: float | int, filter_order: int) -> Recording:
    """Apply IIR bandpass filteer to keep only frequencies above the threshold.

    {SINGLE_FREQ_PARAMS}
    :param filter_order: Controls the strength of filtering. Higher values mean narrower transition region.

    :return: Filtered recording
    """
    _assert_frequency_exists(sample_rate=recording.sample_rate, freq=frequency)
    new_data = recording.data.copy()
    new_data = _bandpass_iir(data=new_data, low=frequency, high=None, sample_rate=recording.sample_rate, filter_order=filter_order)
    return update_data(recording, new_data) 


# Internal


def _notch_fir(data: np.ndarray, frequency: int | float, sample_rate: int | float,
               transition_width: int | float, notch_width: int | float, window: str | None = None) -> np.ndarray:
    low_stop = frequency - notch_width / 2
    high_stop = frequency + notch_width / 2
    numtaps = int(sample_rate / transition_width)
    numtaps = numtaps if numtaps % 2 == 1 else numtaps + 1
    fir_coeff = breesy_scipy._si_firwin_bandstop_multibands(
        numtaps=numtaps,
        bands=[low_stop, high_stop],
        sample_rate=sample_rate,
        window=window
    )
    filtered_data = breesy_scipy._si_filtfilt(b=fir_coeff, a=[1.0], data=data)
    return filtered_data


def _notch_iir(data: np.ndarray, frequency: int | float, sample_rate: int | float, Q: int) -> np.ndarray:
    b, a = breesy_scipy._si_iirnotch(freq=frequency, Q=Q, sample_rate=sample_rate)
    filtered_data = breesy_scipy._si_filtfilt(b=b, a=a, data=data)
    return filtered_data


def _bandpass_fir(data: np.ndarray, low: int | float | None, high: int | float | None, sample_rate: int | float, 
                  transition_width: float, window: str | None = None) -> np.ndarray:

    numtaps = int(sample_rate / transition_width * 3.3)
    numtaps = numtaps if numtaps % 2 == 1 else numtaps + 1

    if low and high:
        fir_coeff = breesy_scipy._si_firwin_bandpass_multibands(
            numtaps=numtaps,
            bands=[low, high],
            sample_rate=sample_rate, 
            window=window
        )
    elif (not low) and high:
        fir_coeff = breesy_scipy._si_firwin_lowpass(numtaps=numtaps, band=high, 
                                                     sample_rate=sample_rate, window=window)
    elif (not high) and low:
        fir_coeff = breesy_scipy._si_firwin_highpass(numtaps=numtaps, band=low, 
                                                     sample_rate=sample_rate, window=window)
    else:
        raise BreesyInternalError(f"At least one frequency should be not None, received: {low=}, {high=}")

    filtered_data = breesy_scipy._si_filtfilt(b=fir_coeff, a=[1.0], data=data)
    return filtered_data


def _bandpass_iir(data: np.ndarray, low: int | float | None, high: int | float | None, sample_rate: int, filter_order: int):
    if low and high:
        the_filter = breesy_scipy._si_butter_bandpass_sos(low=low, high=high, sample_rate=sample_rate, filter_order=filter_order)
    elif (not low) and high:
        the_filter = breesy_scipy._si_butter_lowpass_sos(freq=high, sample_rate=sample_rate, filter_order=filter_order)
    elif (not high) and low:
        the_filter = breesy_scipy._si_butter_highpass_sos(freq=low, sample_rate=sample_rate, filter_order=filter_order)
    else:
        raise BreesyInternalError(f"At least one frequency should be not None, received: {low=}, {high=}")

    return breesy_scipy._si_sosfiltfilt(sos=the_filter, data=data)


def _get_fir_params_from_smoothness(smoothness_factor: int | float, notch_freq: int | float) -> tuple[int | float, int | float]:
    """Uses empirical scaling suggested by Claude Opus 4.5, needs rechecking."""
    bandwidth_3db = notch_freq * smoothness_factor
    notch_width = max(bandwidth_3db * 0.65, 0.1)
    transition_width = max(bandwidth_3db * 0.5, 0.1)
    return round(transition_width, 3), round(notch_width, 3)


def _assert_bandpass_low_high(sample_rate: int | float, low: float | int, high: float | int) -> None:
    low_high_range_valid = low <= high
    if not low_high_range_valid:
        raise BreesyError("Low frequency bound is higher than the high frequency bound",
                          "Provide a low frequency bound value that is less than or equal to the high frequency bound")
    high_bound_exceeds_nyquist = high >= (sample_rate / 2)
    if high_bound_exceeds_nyquist:
        raise BreesyError("High frequency bound exceeds Nyquist frequency",
                          "Provide a high frequency bound value that is less than half of the sample rate")
    if (high <= 0) or (low <= 0):
        raise BreesyError("One or both frequency bounds are not positive",
                          "Ensure both frequency bounds are positive numbers")


def _assert_frequency_exists(sample_rate: int | float, freq: float | int) -> None:
    if freq is None or freq <= 0:  # TODO: test if None check is needed
        raise BreesyError("Frequency bound should be a positive value.",
                          "Provide a frequency bound value that is a positive number.")
    freq_exists = (sample_rate // 2) >= freq
    if not freq_exists:
        raise BreesyError("Frequency bound exceeds Nyquist frequency",
                          "Provide a frequency bound value that is less than half of the sample rate")


import numpy as np
from math import gcd
import itertools
from tqdm.auto import tqdm
from scipy.stats import pearsonr

from . import breesy_scipy, constants, Event
from .recording import Recording, update_data, cut_by_sample_range
from .filtering import notch_filter
from .breesy_scipy import _si_welch, _si_decimate, _si_resample_poly
from .breesy_sklearn import _sklearn_fastica
from .errors import BreesyError, protect_from_lib_error
from .type_hints import enforce_type_hints
from .log import logger


FREQ_SPECTRUM_PARAMS = """
:param scaling: Scaling to use, either "density" (default) for power spectral density, or "spectrum" for squared magnitude spectrum
:param window: Window type to use (see https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.windows.get_window.html)
"""


_DOC_PARAMS = {
    name: value for name, value in globals().items()
    if name.endswith('_PARAMS') and isinstance(value, str)
}


def _format_doc(func):
    """Decorator that expands shared parameter docs in the docstring."""
    if func.__doc__:
        func.__doc__ = func.__doc__.format(**_DOC_PARAMS)
    return func


@_format_doc
@enforce_type_hints
def get_frequency_spectrum_channel(recording: Recording, channel_name: str, scaling: str = 'density',
                                   window: str | None = None) -> tuple[np.ndarray, np.ndarray]:
    """Compute the frequency spectrum of a single EEG channel.

    :param recording: Input EEG recording
    :param channel_name: Name of the channel to process
    {FREQ_SPECTRUM_PARAMS}
    :return: Tuple of (frequencies, power_spectrum)
    """
    if channel_name not in recording.channel_names:
        raise BreesyError(
            f"Channel '{channel_name}' not found in recording.",
            "Use one of the channels in the recording; you can view them by doing: `recording.channel_names`"
        )

    channel_index = recording.channel_names.index(channel_name)
    channel_data = recording.data[channel_index]

    f, spectrum = _si_welch(data=channel_data[np.newaxis, :], sample_rate=recording.sample_rate,
                            scaling=scaling, window=window, nperseg=4*recording.sample_rate)
    return f, spectrum[0]


@_format_doc
@enforce_type_hints
def get_frequency_spectrum(recording: Recording, scaling: str = 'density', window: str | None = None) -> tuple[np.ndarray, np.ndarray]:
    """Compute the frequency spectra of each EEG channel.

    :param recording: Input EEG recording
    {FREQ_SPECTRUM_PARAMS}
    :return: Tuple containing frequencies array and periodogram array of shape (n_channels, n_frequencies)
    """

    f, spectrum = _si_welch(data=recording.data, sample_rate=recording.sample_rate, scaling=scaling, window=window,
                            nperseg=4*recording.sample_rate)
    return f, spectrum


@_format_doc
@enforce_type_hints
def get_mean_frequency_spectrum(recording: Recording, scaling: str = 'density',
                                window: str | None = None) -> tuple[np.ndarray, np.ndarray]:
    """Compute the mean frequency spectrum of all EEG channels.

    :param recording: Input EEG recording
    {FREQ_SPECTRUM_PARAMS}
    :return: Tuple containing frequencies array and periodogram array of shape (n_channels, n_frequencies)
    """

    f, spectrum = get_frequency_spectrum(recording=recording, scaling=scaling, window=window)
    return f, spectrum.mean(axis=0)


# Detection functions --------------------------


@enforce_type_hints
def check_for_data_problems(recording: Recording) -> dict[str, list]:
    """Detect common problems in data:
    - Missing values (NaNs) - if any found, it is highly recommended to repair the data file first before doing any analysis
    - Data scale - estimated unit of data
    - Dead channels - channels which have very low amplitude and likely disconnected during recording
    - Noisy channels - channels which have very high variance and likely overwhelmed by noise
    - Powerline frequency - if there are 50 Hz or 60Hz peaks in some channels, it is recommended to filter out that frequency in all channels

    :param recording: Input recording

    :return: Dictionary with found parameters and problematic values
    """
    nan_channels = detect_nans(recording=recording)
    data_scale = detect_data_scale(recording=recording)
    dead_channels = detect_dead_channels(recording=recording, std_threshold=constants.DEAD_CHANNEL_SD_THRESHOLD, skip_channels=[])
    noisy_channels = detect_noisy_channels(recording=recording, variance_threshold=constants.NOISY_CHANNEL_VARIANCE_THRESHOLD, skip_channels=dead_channels)
    found_powerline_peaks = detect_powerline_noise(recording=recording,
                                                   power_threshold=constants.POWERLINE_POWER_THRESHOLD,
                                                   baseline_l_freq=constants.BASELINE_FOR_PEAKS_LOW,
                                                   baseline_h_freq=constants.BASELINE_FOR_PEAKS_HIGH,
                                                   skip_channels=dead_channels)
    return {
        'missing_values': nan_channels,
        'scale': data_scale,
        'dead': dead_channels,
        'noisy': noisy_channels,
        'powerline': found_powerline_peaks
    }


@enforce_type_hints
@protect_from_lib_error("numpy")
def detect_dead_channels(recording: Recording, std_threshold: float = constants.DEAD_CHANNEL_SD_THRESHOLD, skip_channels: list[str] = None) -> list[str]:
    """Detect channels with abnormally low variability (possible dead channels).

    :param recording: Input recording
    :param std_threshold: Standard deviation threshold for flagging the channel as dead
    :param skip_channels: List of channel names to skip

    :return: List of channel names suspected to be dead
    """
    sid, eid = int(recording.number_of_samples * 0.05), int(recording.number_of_samples * 0.95)
    cut_recording = cut_by_sample_range(recording=recording, start_index=sid, end_index=eid)
    sds = np.std(cut_recording.data, axis=1)
    median_std = np.median(sds)
    dead_channels = [ch for i, ch in enumerate(recording.channel_names) if sds[i] < median_std * std_threshold]
    dead_channels = _skip_channel_by_names(ch_names=dead_channels, skip_names=skip_channels)
    _output_problematic_channels(problem_channels=dead_channels, all_channels=recording.channel_names, reason="dead channels")
    return dead_channels


@enforce_type_hints
@protect_from_lib_error("numpy")
def detect_noisy_channels(recording: Recording, variance_threshold: int | float = constants.NOISY_CHANNEL_VARIANCE_THRESHOLD, skip_channels: list[str] = None) -> list[str]:
    """Detect channels with abnormally high variance (likely noisy channels).

    :param recording: Input recording
    :param variance_threshold: Variance threshold multiplier for flagging the channel as noisy
    :param skip_channels: List of channel names to skip

    :return: List of channel names suspected to be noisy
    """
    variances = np.var(recording.data, axis=1)
    mean_var = np.mean(variances)
    noisy_channels = [ch for i, ch in enumerate(recording.channel_names) if variances[i] > mean_var * variance_threshold]
    noisy_channels = _skip_channel_by_names(ch_names=noisy_channels, skip_names=skip_channels)
    _output_problematic_channels(problem_channels=noisy_channels, all_channels=recording.channel_names, reason="highly noisy channels")
    return noisy_channels


@enforce_type_hints
@protect_from_lib_error("numpy")
def detect_nans(recording: Recording) -> list[str]:
    n_nans_per_channel = np.isnan(recording.data).sum(axis=1)
    nan_channels = [ch for i, ch in enumerate(recording.channel_names) if n_nans_per_channel[i] > 0]
    _output_problematic_channels(problem_channels=nan_channels, all_channels=recording.channel_names, reason="channels with missing values")
    return nan_channels


@enforce_type_hints
@protect_from_lib_error("numpy")
def detect_data_scale(recording: Recording) -> float:
    """Detect the likely scale of EEG data and return a multiplier to convert to microvolts.
    
    This function analyzes the amplitude range of EEG data to determine if it's in
    volts (V), millivolts (mV), or microvolts (µV), and returns the appropriate
    multiplier to convert the data to microvolts.
    
    Typical EEG amplitudes:
    - In microvolts (µV): 10-100 µV (multiplier: 1.0)
    - In millivolts (mV): 0.01-0.1 mV (multiplier: 1000.0)
    - In volts (V): 0.00001-0.0001 V (multiplier: 1000000.0)
    
    :param recording: Input recording

    :return: Multiplier to convert data to microvolts
    """
    channel_stds = np.std(recording.data, axis=-1)
    typical_amplitude = np.median(channel_stds)
    
    if typical_amplitude > 1.0:
        scale = "microvolts (µV)"
        multiplier = 1.0
    elif typical_amplitude > 0.001:
        scale = "millivolts (mV)"
        multiplier = 1000.0
    else:
        scale = "volts (V)"
        multiplier = 1_000_000.0
    
    logger.info(f"EEG data is likely in {scale} (typical amplitude: {typical_amplitude:.6f})")
    
    return multiplier



def _skip_channel_by_names(ch_names: list[str], skip_names: list[str] | None) -> list[str]:
    if skip_names:
        return [ch for ch in ch_names if ch not in skip_names]
    return ch_names


def _output_problematic_channels(problem_channels: list[str], all_channels: list[str], reason: str) -> None:
    if not problem_channels:
        logger.info(f'Checked for {reason}: no problematic channels detected.')
    elif len(problem_channels) == len(all_channels):
        logger.info(f'Checked for {reason}: detected in all channels.')
    else:
        ratio = round(len(problem_channels) / len(all_channels) * 100)
        ch_names_str = ' - ' + ', '.join(problem_channels) if len(problem_channels) < 10 else ''
        logger.info(f"Checked for {reason}: likely detected in {len(problem_channels)} ({ratio}%) channel(s){ch_names_str}.")


@enforce_type_hints
def detect_powerline_noise(recording: Recording, power_threshold: float = constants.POWERLINE_POWER_THRESHOLD,
                           baseline_l_freq: int | float = constants.BASELINE_FOR_PEAKS_LOW,
                           baseline_h_freq: int | float = constants.BASELINE_FOR_PEAKS_HIGH,
                           skip_channels: list[str] = None) -> list[str]:
    if recording.sample_rate / 2 < 50:
        logger.info(f'Powerline noise cannot be detected because sample rate is too low.')
        return []

    f, spectra = get_frequency_spectrum(recording=recording)
    baseline = _get_baseline_for_peak_detection(f=f, x=spectra, start_freq=baseline_l_freq, end_freq=baseline_h_freq)
    if skip_channels:
        keep_mask = np.array([False if ch in skip_channels else True for ch in recording.channel_names])
    else:
        keep_mask = np.full(recording.number_of_channels, True)

    powerline_peaks, power_freq = None, None

    powerline_freqs = [50, 60]
    for freq in powerline_freqs:
        if recording.sample_rate / 2 < freq:
            break
        if _notch_filter_detected(f=f, x=spectra, freq=freq):
            continue

        peaks = _measure_rel_peaks_at(f=f, x=spectra, freq=freq, baseline=baseline)
        if peaks[keep_mask].max() > power_threshold:
            powerline_peaks = peaks
            power_freq = freq
            break

    if power_freq is not None:
        powerline_peaks[~keep_mask] = 1  # set to baseline
        noisy_channels = [ch for i, ch in enumerate(recording.channel_names)
                          if powerline_peaks[i] > power_threshold]
        _output_problematic_channels(problem_channels=noisy_channels, all_channels=recording.channel_names,
                                     reason=f'prominent {power_freq} Hz powerline noise')
        # harmonics
        noisiest_i = np.argmax(powerline_peaks).ravel()
        for harmonic in range(power_freq*2, int(recording.sample_rate // 2), power_freq):
            harmonic_peak = _measure_rel_peaks_at(f=f, x=spectra[noisiest_i], freq=harmonic, baseline=baseline[noisiest_i])
            if harmonic_peak > power_threshold:
                logger.info(f'\t+ harmonic of {harmonic} Hz')

        return noisy_channels
    else:
        logger.info(f'No powerline noise detected or was already filtered.')
        return []


def _notch_filter_detected(f: np.ndarray, x: np.ndarray, freq: int) -> bool:
    freq_i = np.searchsorted(f, freq)
    power_at_freq = x[:, freq_i]

    window_size = 5
    attenuation_threshold = 0.3
    ch_ratio_with_notch = 0.5
    lower_i = np.searchsorted(f, freq - window_size)
    upper_i = np.searchsorted(f, freq + window_size)
    surrounding_power = np.concatenate([
        x[:, lower_i:freq_i], x[:, freq_i+1:upper_i]
    ], axis=1)
    avg_surrounding = np.mean(surrounding_power, axis=1)
    ratio = power_at_freq / avg_surrounding
    return np.sum(ratio < attenuation_threshold) > x.shape[0] * ch_ratio_with_notch


@protect_from_lib_error("numpy")
def _get_baseline_for_peak_detection(f: np.ndarray, x: np.ndarray, start_freq: int | float, end_freq: int | float) -> np.ndarray:
   start_i = np.searchsorted(f, start_freq)
   end_i = np.searchsorted(f, end_freq)
   baseline = np.percentile(x[:, start_i:end_i], q=95, axis=1)
   return baseline


@protect_from_lib_error("numpy")
def _measure_rel_peaks_at(f: np.ndarray, x: np.ndarray, freq: int | float, baseline: float) -> np.ndarray:
    freq_i = np.searchsorted(f, freq)
    return x[:, freq_i] / baseline


# ----------------

# TODO: move to a new file called decomposition? or to transform?
@enforce_type_hints
def get_ica_components(recording: Recording, n_components: int, random_state: int | None = None) -> np.ndarray:
    """Compute Independent Component Analysis (ICA) components.

    :param recording: Input recording with data shaped (n_channels, n_samples)
    :param n_components: Number of ICA components to compute
    :param random_state: Optional random seed for reproducibility

    :return: Components array of shape (n_components, n_samples)
    """
    ica = _sklearn_fastica(n_components=n_components, random_state=random_state)
    components = ica.fit_transform(recording.data.T)
    return components.T


@enforce_type_hints
def get_channel_correlation(recording: Recording, absolute_values: bool = True) -> np.ndarray:
    """Compute matrix of Pearson correlation between all channels in data.

    :param recording: Input EEG recording
    :absolute_values: Whether to ues absolute (in [0, 1] range) correlation values. Setting this False will keep correlation in [-1, 1] range. 

    :return: Correlation array of shape (n_channels, n_channels)
    """
    n_channels = recording.number_of_channels
    matrix = np.ones((n_channels, n_channels))
    matrix_iter = tqdm(itertools.combinations(range(n_channels), r=2), desc="Calculating correlations between channels",
                       total=int(n_channels * (n_channels-1) / 2), disable=n_channels < 33)
    for i, j in matrix_iter:
        corr = pearsonr(recording.data[i], recording.data[j]).statistic
        if absolute_values:
            corr = abs(corr)
        matrix[i, j] = corr
        matrix[j, i] = corr
    return matrix

# ----------------


@enforce_type_hints
def remove_powerline_noise(recording: Recording, smoothness_factor: int | float = constants.NOTCH_DEFAULT_SMOOTHNESS,
                           power_threshold: float = constants.POWERLINE_POWER_THRESHOLD) -> Recording:
    """Remove powerline noise from a recording.

    :param recording: Input recording
    :param quality_factor: Quality factor for the notch filter

    :return: Filtered recording
    """
    frequency = detect_powerline_noise(recording=recording, power_threshold=power_threshold)
    if frequency:
        for freq in frequency:
            recording = notch_filter(recording=recording, frequency=freq, smoothness_factor=smoothness_factor)
        return recording
    return recording


@enforce_type_hints
@protect_from_lib_error("numpy")
def mean_centering(recording: Recording) -> Recording:
    """Center the recording data by subtracting the mean.

    :param recording: Input recording

    :return: Mean-centered recording
    """
    new_eeg_data = recording.data - np.expand_dims(recording.data.mean(axis=-1), 1)
    return update_data(recording, new_eeg_data)


@enforce_type_hints
def remove_slow_drift(recording: Recording, highpass_freq: int | float = constants.SLOWDRIFT_REMOVAL_FREQ,
                      filter_order: int = constants.BUTTER_FILTER_ORDER) -> Recording:
    """Remove slow drift using a high-pass Butterworth filter.

    :param recording: Input recording
    :param highpass_freq: High-pass cutoff frequency in Hz
    :param filter_order: Order of the Butterworth high-pass filter

    :return: Filtered recording
    """
    highpass = breesy_scipy._si_butter_highpass_sos(freq=highpass_freq, sample_rate=recording.sample_rate, filter_order=filter_order)
    new_eeg_data = breesy_scipy._si_sosfiltfilt(sos=highpass, data=recording.data)
    return update_data(recording, new_eeg_data)


@enforce_type_hints
@protect_from_lib_error("numpy")
def rereference(recording: Recording, channel_name: str) -> Recording:
    """Re-reference EEG by subtracting a reference signal from each channel.

    :param recording: Input recording
    :param channel_name: Name of the reference signal to subtract

    :return: Re-referenced recording
    """
    ch_i = recording.channel_names.find(channel_name)
    if ch_i == -1:
        available_channels = ", ".join([f'"{x}"' for x in recording.channel_names])
        raise BreesyError(f'Channel "{channel_name}" was not found in recording.',
                          f'Use one of available channels: {available_channels}.')
    new_eeg_data = recording.data - recording.data[ch_i]
    return update_data(recording, new_eeg_data)


@enforce_type_hints
def downsample(recording: Recording, new_sample_rate: int | float) -> Recording:
    """Downsample the recording to a new sample rate.
    
    :param recording: Input recording
    :param new_sample_rate: Target sample rate (must be lower than current sample rate)
    
    :return: Downsampled recording
    """
    if new_sample_rate >= recording.sample_rate:
        raise BreesyError(f'New sample rate ({new_sample_rate} Hz) must be lower than the current sample rate ({recording.sample_rate} Hz).',
                          "Provide a lower sample rate to downsample the data.")

    downsample_factor = recording.sample_rate / new_sample_rate
    if downsample_factor.is_integer():
        new_data = _si_decimate(recording.data, int(downsample_factor))
    else:
        g = gcd(int(new_sample_rate), int(recording.sample_rate))
        up = int(new_sample_rate / g)
        down = int(recording.sample_rate / g)
        new_data = _si_resample_poly(
            recording.data,
            up=up,
            down=down,
            axis=-1)

    new_events = [
        Event(event.name, index=int(event.index / downsample_factor))
        for event in recording.events
    ]
    return Recording(new_data, recording.channel_names, new_sample_rate, new_events)



from pathlib import Path
import numpy as np

from .recording import Recording
from .Event import Event
from .RecordingMetadata import RecordingMetadata
from .breesy_scipy import _loadmat
from .errors import BreesyError, BreesyInternalError
from .load import _leave_only_eeg_channels
from .type_hints import enforce_type_hints


@enforce_type_hints
def load_eeglab(filename: str) -> Recording:
    """Load an EEGlab .set file (with optional .fdt companion file).

    :param filename: Path to the .set or .fdt file

    :return: Recording object containing the loaded data
    """
    if filename.endswith('.fdt'):
        filename = Path(filename).with_suffix('.set')
        if not filename.is_file():
            raise BreesyError('".fdt" filename provided, but it cannot be read without a corresponding ".set" file, which was not found.',
                              'Ensure that the ".set" file is in the same directory as your ".fdt" file. Their names should be the same, except for the extension.')

    set_data = _loadmat(filename)

    eeg = _extract_eeg_structure(set_data, filename)
    _validate_continuous_data(eeg, filename)

    # sample rate, channels, dimensions
    sample_rate, nbchan, pnts = _extract_metadata(eeg, filename)
    channel_names = _extract_channel_names(eeg, nbchan, filename)

    data = _load_eeglab_data(eeg, filename, nbchan, pnts)
    data, channel_names = _leave_only_eeg_channels(data, channel_names)
    events = _extract_events(eeg, sample_rate)

    metadata = RecordingMetadata(
        name=Path(filename).stem,
        file_path=Path(filename),
        description=f"EEGlab data loaded from {filename}"
    )

    return Recording(
        data=data,
        channel_names=channel_names,
        sample_rate=sample_rate,
        events=events,
        metadata=metadata
    )


def _extract_eeg_structure(set_data: dict, filename: str) -> dict:
    # Most EEGlab files have a top-level 'EEG' key
    if 'EEG' in set_data:
        eeg_raw = set_data['EEG']
        # Convert structured array to dict for easier access
        return _struct_to_dict(eeg_raw)

    # Fallback: look for meaningful keys (not MATLAB metadata)
    meaningful_keys = [k for k in set_data.keys() if not k.startswith('__')]

    if len(meaningful_keys) == 0:
        raise BreesyError(
            "No valid data found in .set file.",
            f"The file {filename} appears to be empty or corrupted. "
            "Please check that this is a valid EEGlab .set file."
        )

    if len(meaningful_keys) == 1:
        return _struct_to_dict(set_data[meaningful_keys[0]])

    eeg_like_keys = [k for k in meaningful_keys
                     if any(marker in k.lower() for marker in ['eeg', 'data', 'set'])]

    if len(eeg_like_keys) == 1:
        return _struct_to_dict(set_data[eeg_like_keys[0]])

    raise BreesyError(
        f"Could not identify EEG structure in .set file. Found multiple possible keys: {meaningful_keys}",
        "The file structure is ambiguous. Please ensure this is a valid EEGlab .set file."
    )


def _struct_to_dict(struct) -> dict:
    if isinstance(struct, dict):
        return struct

    # Numpy structured array (record array)
    if isinstance(struct, np.ndarray) and hasattr(struct, 'dtype') and struct.dtype.names:
        result = {}
        for field_name in struct.dtype.names:
            result[field_name] = struct[field_name]
        return result

    # Not a struct - return as-is wrapped in a dict
    return {'data': struct}


def _validate_continuous_data(eeg: dict, filename: str) -> None:
    # Check if this is epoched data
    trials = eeg.get('trials', 1)

    # Handle different data types that scipy.io.loadmat might return
    if isinstance(trials, np.ndarray):
        trials = int(trials.flat[0])
    else:
        trials = int(trials)

    if trials != 1:
        raise BreesyError(
            f"This EEGlab file contains epoched data ({trials} epochs/trials). "
            "Breesy currently supports only continuous EEG data.",
            "Please export your data as continuous from EEGlab before loading, "
            "or merge your epochs back into continuous data."
        )


def _extract_metadata(eeg: dict, filename: str) -> tuple[float, int, int]:
    # Extract sample rate (srate)
    if 'srate' not in eeg:
        raise BreesyError(
            "Missing 'srate' (sampling rate) field in .set file.",
            f"The file {filename} does not contain sampling rate information. "
            "Please check that this is a valid EEGlab .set file."
        )
    sample_rate = _extract_scalar_value(eeg['srate'], 'srate', filename)
    sample_rate = float(sample_rate)

    if sample_rate <= 0:
        raise BreesyError(
            f"Invalid sample rate: {sample_rate} Hz",
            "Sample rate must be positive. Please check the .set file."
        )

    # Extract number of channels (nbchan)
    if 'nbchan' not in eeg:
        raise BreesyError(
            "Missing 'nbchan' (number of channels) field in .set file.",
            f"The file {filename} does not contain channel count information. "
            "Please check that this is a valid EEGlab .set file."
        )
    nbchan = _extract_scalar_value(eeg['nbchan'], 'nbchan', filename)
    nbchan = int(nbchan)

    if nbchan <= 0:
        raise BreesyError(
            f"Invalid number of channels: {nbchan}",
            "Number of channels must be positive. Please check the .set file."
        )

    # Extract number of time points (pnts)
    if 'pnts' not in eeg:
        raise BreesyError(
            "Missing 'pnts' (number of time points) field in .set file.",
            f"The file {filename} does not contain time points information. "
            "Please check that this is a valid EEGlab .set file."
        )
    pnts = _extract_scalar_value(eeg['pnts'], 'pnts', filename)
    pnts = int(pnts)

    if pnts <= 0:
        raise BreesyError(
            f"Invalid number of time points: {pnts}",
            "Number of time points must be positive. Please check the .set file."
        )

    return sample_rate, nbchan, pnts


def _extract_channel_names(eeg: dict, nbchan: int, filename: str) -> list[str]:
    default_names = [f'Channel_{i+1}' for i in range(nbchan)]

    if 'chanlocs' not in eeg:
        return default_names

    chanlocs = eeg['chanlocs']
    if chanlocs is None or (isinstance(chanlocs, np.ndarray) and chanlocs.size == 0):
        return default_names

    try:
        labels = _extract_labels_from_chanlocs(chanlocs, nbchan)
        if labels and len(labels) == nbchan:
            return labels
    except Exception:
        pass

    return default_names


def _extract_labels_from_chanlocs(chanlocs, nbchan: int) -> list[str] | None:
    # Handle scalar numpy arrays (common with scipy.io.loadmat)
    if isinstance(chanlocs, np.ndarray) and chanlocs.shape == ():
        chanlocs = chanlocs.item()

    # Case 1: chanlocs is a structured numpy array (most common)
    if isinstance(chanlocs, np.ndarray):
        # Check if it has 'labels' field
        if hasattr(chanlocs, 'dtype') and chanlocs.dtype.names:
            if 'labels' in chanlocs.dtype.names:
                labels = chanlocs['labels']
                return _normalize_labels(labels, nbchan)

        # Check for alternative field names
        for field_name in ['label', 'name', 'channel']:
            if hasattr(chanlocs, 'dtype') and chanlocs.dtype.names and field_name in chanlocs.dtype.names:
                labels = chanlocs[field_name]
                return _normalize_labels(labels, nbchan)

    # Case 2: chanlocs is a list or dict
    if isinstance(chanlocs, (list, tuple)):
        try:
            labels = [ch.get('labels', ch.get('label', '')) if isinstance(ch, dict) else str(ch)
                     for ch in chanlocs]
            if len(labels) == nbchan:
                return [str(label).strip() for label in labels if label]
        except Exception:
            pass

    return None


def _normalize_labels(labels, nbchan: int) -> list[str] | None:
    try:
        # Convert to list if numpy array
        if isinstance(labels, np.ndarray):
            # Flatten and convert to strings
            if labels.ndim == 0:
                # Scalar - single channel
                return [str(labels.item()).strip()]
            elif labels.ndim == 1:
                # 1D array - multiple channels
                result = [str(label).strip() for label in labels]
                if len(result) == nbchan and all(result):
                    return result
            else:
                # Multi-dimensional - try to flatten
                result = [str(label).strip() for label in labels.flat]
                if len(result) == nbchan and all(result):
                    return result

        # Already a list or tuple
        elif isinstance(labels, (list, tuple)):
            result = [str(label).strip() for label in labels]
            if len(result) == nbchan and all(result):
                return result

    except Exception:
        pass

    return None


def _load_eeglab_data(eeg: dict, set_filename: str, nbchan: int, pnts: int) -> np.ndarray:
    """Load EEG data from embedded array or external .fdt file.

    EEGlab can store data in two ways:
    1. Embedded in the .set file (as a numpy array in the 'data' field)
    2. In an external binary .fdt file (referenced by filename in 'data' field)
    """
    if 'data' not in eeg:
        raise BreesyError(
            "Missing 'data' field in .set file.",
            f"The file {set_filename} does not contain EEG data. "
            "Please check that this is a valid EEGlab .set file."
        )

    data_field = eeg['data']

    # Handle scalar numpy arrays (common with scipy.io.loadmat)
    if isinstance(data_field, np.ndarray) and data_field.shape == ():
        data_field = data_field.item()

    # Case 1: Data is in an external .fdt file (referenced by filename string)
    if isinstance(data_field, str):
        fdt_filename = data_field
        return _load_fdt_data(fdt_filename, set_filename, nbchan, pnts)

    # Case 2: Data is embedded in the .set file as a numpy array
    if isinstance(data_field, np.ndarray):
        return _process_embedded_data(data_field, eeg, nbchan, pnts, set_filename)

    # Case 3: Unexpected data type
    raise BreesyInternalError(
        f"Unexpected data type in 'data' field: {type(data_field)} "
        f"The .set file {set_filename} has an unexpected structure. "
        "Expected numpy array or filename string."
    )


def _process_embedded_data(data: np.ndarray, eeg: dict, nbchan: int, pnts: int, filename: str) -> np.ndarray:
    # Validate data shape
    expected_shape = (nbchan, pnts)

    if data.shape != expected_shape:
        # Try transposed shape
        if data.shape == (pnts, nbchan):
            data = data.T
        else:
            raise BreesyInternalError(
                f"Embedded data has unexpected shape: {data.shape}. Expected {expected_shape} "
                f"The .set file {filename} has inconsistent metadata. "
                f"Expected {nbchan} channels and {pnts} time points."
            )

    # Apply calibration scaling if present
    # EEGlab sometimes stores a 'CAL' field for unit conversion (typically 1e-6 for microvolts)
    if 'CAL' in eeg:
        try:
            cal = _extract_scalar_value(eeg['CAL'], 'CAL', filename)
            data = data * float(cal)
        except Exception:
            # If calibration fails, just use raw data
            pass

    return data


def _load_fdt_data(fdt_filename: str, set_filename: str, nbchan: int, pnts: int) -> np.ndarray:
    # Resolve .fdt file path
    # The fdt_filename can be:
    # 1. Just a filename (e.g., "data.fdt") - look in same directory as .set
    # 2. A relative path (e.g., "../data/data.fdt") - resolve relative to .set
    # 3. An absolute path (e.g., "/full/path/data.fdt") - use as-is

    set_path = Path(set_filename)
    fdt_path = Path(fdt_filename)

    if not fdt_path.is_absolute():
        # Relative path or just filename - resolve relative to .set file's directory
        fdt_path = set_path.parent / fdt_path

    # If still not found, try replacing .set extension with .fdt
    if not fdt_path.exists():
        fdt_path = set_path.with_suffix('.fdt')

    # Final check - does the file exist?
    if not fdt_path.exists():
        raise BreesyError(
            f"Data file not found: {fdt_filename}",
            f"EEGlab .set files require a corresponding .fdt file. "
            f"Looked in: {fdt_path.parent}\n"
            f"Make sure the .fdt file is in the same directory as the .set file."
        )

    # Read the binary .fdt file
    return _read_fdt_file(str(fdt_path), nbchan, pnts)


def _read_fdt_file(fdt_path: str, nbchan: int, pnts: int) -> np.ndarray:
    """Read binary .fdt file and reshape to (channels, samples).

    .fdt files store data as:
    - Data type: 32-bit IEEE floating point (float32)
    - Byte order: Little-endian
    - Memory order: Fortran/column-major (samples fastest)
    """
    try:
        # Read raw binary data as float32
        data = np.fromfile(fdt_path, dtype=np.float32)

        # Validate size
        expected_size = nbchan * pnts
        if data.size != expected_size:
            raise BreesyError(
                f"Data size mismatch in .fdt file. "
                f"Expected {expected_size} values ({nbchan} channels × {pnts} samples), "
                f"but got {data.size} values.",
                "The .fdt file may be corrupted or does not match the .set metadata. "
                "Make sure the .fdt and .set files are from the same recording."
            )

        # Reshape from flat array to 2D
        # .fdt files are in Fortran (column-major) order:
        # - Samples vary fastest (time points in each channel are contiguous)
        # - Channels vary slowest
        # This means the file is organized as: [ch1_all_samples, ch2_all_samples, ...]
        # So we reshape as (pnts, nbchan) with Fortran order, then transpose
        data = data.reshape((pnts, nbchan), order='F').T

        return data

    except BreesyError:
        # Re-raise our own errors
        raise
    except PermissionError:
        raise BreesyError(
            f"Permission denied when reading .fdt file: {fdt_path}",
            "Check that you have read permissions for this file."
        )
    except Exception as e:
        raise BreesyInternalError(
            f"Failed to read .fdt file: {str(e)} "
            f"The file {fdt_path} may be corrupted or in an unexpected format."
        )


def _extract_events(eeg: dict, sample_rate: float) -> list[Event]:
    """Extract event markers from EEG structure.

    EEGlab stores events in the 'event' field as a structured array with fields like:
    - 'type': Event type/name
    - 'latency': Sample index (1-based in EEGlab)
    - 'duration': Event duration (optional)

    :return: List of Event objects sorted by index
    """
    # Check if events field exists
    if 'event' not in eeg:
        return []

    events_data = eeg['event']

    # Handle empty events
    if events_data is None:
        return []

    # Handle case where it's not an array
    if not isinstance(events_data, np.ndarray):
        return []

    # Handle empty array
    if events_data.size == 0:
        return []

    # Unwrap 0-d object arrays
    if events_data.dtype == object and events_data.shape == ():
        events_data = events_data.item()
        if not isinstance(events_data, np.ndarray):
            return []

    # Extract events from structured array
    events = []

    try:
        # Convert to 1D array if needed
        events_array = np.atleast_1d(events_data)

        for event_item in events_array:
            try:
                event_obj = _parse_single_event(event_item)
                if event_obj:
                    events.append(event_obj)
            except Exception:
                # Skip malformed events
                continue

    except Exception:
        # If anything goes wrong, return empty list
        return []

    # Sort events by index
    events.sort(key=lambda e: e.index)

    return events


def _parse_single_event(event_item) -> Event | None:
    try:
        # Extract event type/name and latency (sample index)
        event_type, latency = None, None

        if hasattr(event_item, 'dtype') and event_item.dtype.names:
            if 'type' in event_item.dtype.names:
                event_type = event_item['type']
            if 'latency' in event_item.dtype.names:
                latency = event_item['latency']
        elif isinstance(event_item, dict):
            event_type = event_item.get('type')
            latency = event_item.get('latency')

        # no latency = no sample index = no event
        if latency is None:
            return None
        if isinstance(latency, np.ndarray):
            if latency.size == 0:
                return None
            latency_value = int(latency.flat[0])
        else:
            latency_value = int(latency)
        # EEGlab uses 1-based indexing, convert to 0-based
        sample_index = latency_value - 1
        if sample_index < 0:
            return None    

        # no event type = default event
        if event_type is None:
            event_name = 'Event'
        elif isinstance(event_type, np.ndarray):
            event_name = str(event_type.flat[0]) if event_type.size > 0 else 'Event'
        elif isinstance(event_type, (bytes, np.bytes_)):
            event_name = event_type.decode('utf-8', errors='ignore')
        else:
            event_name = str(event_type)

        event_name = event_name.strip()
        if not event_name:
            event_name = 'Event'

        return Event(name=event_name, index=sample_index)

    except Exception:
        return None


def _extract_scalar_value(value, field_name: str, filename: str):
    if isinstance(value, (int, float)):
        return value
    if isinstance(value, np.ndarray):
        if value.size == 1:
            return value.flat[0]
        else:
            raise BreesyError(
                f"Expected scalar value for '{field_name}', but got array of size {value.size}",
                f"The .set file {filename} has an unexpected structure."
            )
    if isinstance(value, (list, tuple)) and len(value) == 1:
        return value[0]

    raise BreesyInternalError(
        f"Could not extract scalar value from '{field_name}' field. Got type: {type(value)} "
        f"The .set file {filename} has an unexpected structure."
    )



from pathlib import Path
import numpy as np

from breesy import Event, recording
from breesy.load import load_recording
from breesy.processing import get_frequency_spectrum_channel, get_frequency_spectrum, downsample

REPO_ROOT = Path(__file__).parent.parent


def test_get_frequency_spectrum_channel():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    rec = load_recording(str(datapath))
    first_channel_name = rec.channel_names[0]

    # Act
    f, spectrum = get_frequency_spectrum_channel(rec, first_channel_name)

    # Assert
    assert f.ndim == 1
    assert spectrum.ndim == 1
    assert len(f) == len(spectrum)
    assert np.all(f >= 0)
    assert np.all(spectrum > 0)
    assert np.all(np.isfinite(spectrum))


def test_get_frequency_spectrum_channel_matches_full():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    rec = load_recording(str(datapath))
    first_channel_name = rec.channel_names[0]

    # Act
    f_full, spectrum_full = get_frequency_spectrum(rec)
    f_single, spectrum_single = get_frequency_spectrum_channel(rec, first_channel_name)

    # Assert
    assert np.array_equal(f_full, f_single)
    assert np.allclose(spectrum_full[0], spectrum_single)

def test_downsample_integer_factor():
    # Arrange
    original_rate = 512
    target_rate = 128  # factor of 4 (0.25)

    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    rec = load_recording(str(datapath), sample_rate=original_rate)
    
    test_events = [
        Event(name="start", index=512),  # Expect at 128
        Event(name="middle", index=1024), # Expect at 256
        Event(name="end", index=2048) # Expect at 512
    ]
    rec = recording.add_events(rec, test_events)
    
    # Act
    downsampled = downsample(rec, new_sample_rate=target_rate)
    
    # Assert
    assert downsampled.sample_rate == target_rate
    assert downsampled.channel_names == rec.channel_names
    assert downsampled.number_of_channels == rec.number_of_channels
    
    assert downsampled.number_of_samples == rec.number_of_samples / 4
    
    assert not np.array_equal(downsampled.data[:, :50], rec.data[:, :50])
    
    # Events
    assert len(downsampled.events) == len(test_events)

    assert downsampled.events[0].index == 128
    assert downsampled.events[1].index == 256
    assert downsampled.events[2].index == 512


def test_downsample_float_factor():
    # Arrange
    original_rate = 512
    target_rate = 384  # factor of 1.33(3) (0.75)
    
    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    rec = load_recording(str(datapath), sample_rate=original_rate)
    
    test_events = [
        Event(name="start", index=512),  # Expect at 384
        Event(name="middle", index=1024),  # Expect at 768
        Event(name="end", index=2048)  # Expect at 1536
    ]
    rec = recording.add_events(rec, test_events)
    
    # Act
    downsampled = downsample(rec, new_sample_rate=target_rate)
    
    # Assert
    assert downsampled.sample_rate == target_rate
    assert downsampled.channel_names == rec.channel_names
    assert downsampled.number_of_channels == rec.number_of_channels
    
    assert downsampled.number_of_samples == rec.number_of_samples * 3/4
    
    assert not np.array_equal(downsampled.data[:, :50], rec.data[:, :50])
    
    # Events
    assert len(downsampled.events) == len(test_events)
    
    assert downsampled.events[0].index == 384
    assert downsampled.events[1].index == 768
    assert downsampled.events[2].index == 1536


from pathlib import Path
import struct
import numpy as np
from scipy.stats import mode
from typing import BinaryIO

from .load import _is_likely_eeg_column
from .recording import Recording
from .RecordingMetadata import RecordingMetadata
from .errors import BreesyError, BreesyInternalError
from .log import logger
from .type_hints import enforce_type_hints


@enforce_type_hints
def load_edf(filename: str) -> Recording:
    """Load an EDF (European Data Format) or EDF+ file.

    :param filename: Path to the EDF file

    :return: Recording object containing the loaded data
    """
    with open(filename, 'rb') as f:
        # assert it is EDF
        first_byte = f.read(1)
        f.seek(0)
        if first_byte != b"0":
            raise BreesyError(f'The file is not correctly formatted for an EDF file.',
                              f'Ensure that "{filename}" is a proper EDF(+) file. Consider contacting authors of the file.')

        data, channel_names, sample_rate = _read_edf_or_bdf(f, bytes_per_sample=2)


    metadata = RecordingMetadata(
        name=Path(filename).stem,
        file_path=Path(filename),
        description=f"EDF data loaded from {filename}"
    )

    return Recording(
        data=data,
        channel_names=channel_names,
        sample_rate=sample_rate,
        metadata=metadata
    )


@enforce_type_hints
def load_bdf(filename: str) -> Recording:
    """Load a BDF (BioSemi Data Format) or BDF+ file.

    :param filename: Path to the BDF file

    :return: Recording object containing the loaded data
    """        
    with open(filename, 'rb') as f:
        # assert it is BDF
        first_byte = f.read(1)
        f.seek(0)
        if first_byte[0] != 255:
            raise BreesyError(f'The file is not correctly formatted for an BDF file.',
                              f'Ensure that "{filename}" is a proper BDF(+) file. Consider contacting authors of the file.')

        data, channel_names, sample_rate = _read_edf_or_bdf(f, bytes_per_sample=3)

    metadata = RecordingMetadata(
        name=Path(filename).stem,
        file_path=Path(filename),
        description=f"BDF data loaded from {filename}"
    )

    return Recording(
        data=data,
        channel_names=channel_names,
        sample_rate=sample_rate,
        metadata=metadata
    )


def _read_header(f: BinaryIO) -> dict:
    # Read fixed header (256 bytes)
    header = {}
    header['version'] = f.read(8).decode('ascii', errors='ignore').strip()
    header['patient_id'] = f.read(80).decode('ascii').strip()
    header['recording_id'] = f.read(80).decode('ascii').strip()
    header['start_date'] = f.read(8).decode('ascii').strip()
    header['start_time'] = f.read(8).decode('ascii').strip()
    header['header_bytes'] = int(f.read(8).decode('ascii').strip())
    header['reserved'] = f.read(44).decode('ascii').strip()
    header['num_records'] = int(f.read(8).decode('ascii').strip())
    header['record_duration'] = float(f.read(8).decode('ascii').strip())
    header['num_signals'] = int(f.read(4).decode('ascii').strip())
    return header


def _read_signal_header(f: BinaryIO, num_signals: int) -> dict:
    # Read signal headers (256 bytes per signal)
    signal_headers = {}
    signal_headers['label'] = [f.read(16).decode('ascii').strip() for _ in range(num_signals)]
    signal_headers['transducer'] = [f.read(80).decode('ascii').strip() for _ in range(num_signals)]
    signal_headers['physical_dim'] = [f.read(8).decode('ascii').strip() for _ in range(num_signals)]
    signal_headers['physical_min'] = [float(f.read(8).decode('ascii').strip()) for _ in range(num_signals)]
    signal_headers['physical_max'] = [float(f.read(8).decode('ascii').strip()) for _ in range(num_signals)]
    signal_headers['digital_min'] = [int(f.read(8).decode('ascii').strip()) for _ in range(num_signals)]
    signal_headers['digital_max'] = [int(f.read(8).decode('ascii').strip()) for _ in range(num_signals)]
    signal_headers['prefilter'] = [f.read(80).decode('ascii').strip() for _ in range(num_signals)]
    signal_headers['samples_per_record'] = [int(f.read(8).decode('ascii').strip()) for _ in range(num_signals)]
    signal_headers['reserved'] = [f.read(32) for _ in range(num_signals)]
    return signal_headers


def _get_scaling_factors_and_offsets(signal_headers: dict, num_signals: int) -> tuple[list, list]:
    scaling_factors, offsets = [], []
    for i in range(num_signals):
        digital_range = signal_headers['digital_max'][i] - signal_headers['digital_min'][i]
        physical_range = signal_headers['physical_max'][i] - signal_headers['physical_min'][i]
        if digital_range != 0:
            scaling_factors.append(physical_range / digital_range)
            offsets.append(signal_headers['physical_min'][i] - signal_headers['digital_min'][i] * scaling_factors[i])
        else:
            scaling_factors.append(1.0)
            offsets.append(0.0)
    return scaling_factors, offsets


def _filter_plus_annotations(data: list[np.ndarray], channel_names: list[str]) -> tuple[list[np.ndarray], list[str], list[int]]:
    annot_labels = ['EDF Annotations', 'BDF Annotations']
    other_labels = ['Status']
    likely_eeg = [_is_likely_eeg_column(ch) for ch in data]
    eeg_indices = [i for i, label in enumerate(channel_names)
                   if all(x not in label for x in annot_labels+other_labels)
                   and likely_eeg[i]]
    removed = [name for i, name in enumerate(channel_names) if i not in eeg_indices]
    if removed:
        logger.warning(f"Ignored {len(removed)} channel(s) that don't look like EEG signal: {removed}")
        data = [data[i] for i in eeg_indices]
        channel_names = [channel_names[i] for i in eeg_indices]
        # TODO: create events from annotations
    return data, channel_names, eeg_indices


def _determine_sample_rate(samples_per_record: list[int], header: dict, eeg_indices: list[int], channel_names: list[str]) -> int:
    sample_rates = [samples_per_record[i] / header['record_duration'] for i in eeg_indices]
    if len(set(sample_rates)) == 1:
        sample_rate = float(sample_rates[0])
    else:
        sample_rate = mode(sample_rates)
        diff_channels_str = ', '.join([ch for i, ch in enumerate(channel_names) if sample_rates[i] != sample_rate])
        logger.warning(f'Detected variable sample rates. Will use {sample_rate} as the most common one. Channel names with different sample rates: {diff_channels_str}')
    return sample_rate


def _convert_to_microvolts(data: list[np.ndarray], units: list[str]) -> list[np.ndarray]:
    converted_data = []
    for i, (channel_data, unit) in enumerate(zip(data, units)):
        unit_normalized = unit.lower().strip()
        
        if unit_normalized in ['μv', 'uv', 'µv']:
            conversion_factor = 1.0
        elif unit_normalized == 'mv':
            conversion_factor = 1000.0
        elif unit_normalized == 'v':
            conversion_factor = 1_000_000.0
        else:
            raise BreesyError(f'Unknown unit specified for channel nr. {i+1}: "{unit}".',
                              f'Contact the authors of the file and ask to fix it. Breesy only supports "μV", "mV", and "V" (not case-sensitive). '
                              f'If you think that "{unit}" unit should also be supported, contact Breesy authors.')
        
        converted_data.append(channel_data * conversion_factor)
    
    return converted_data


def _read_edf_or_bdf(f: BinaryIO, bytes_per_sample: int) -> tuple[np.ndarray, list[str], int]:
    if bytes_per_sample not in [2, 3]:
        raise BreesyInternalError(f'bytes_per_sample can only be 2 or 3, provided {bytes_per_sample}.')
    
    header = _read_header(f)
    num_signals = header['num_signals']
    signal_headers = _read_signal_header(f, num_signals=num_signals)
    scaling_factors, offsets = _get_scaling_factors_and_offsets(signal_headers=signal_headers, num_signals=num_signals)

    num_records = header['num_records']
    samples_per_record = signal_headers['samples_per_record']
    total_samples = [samples_per_record[i] * num_records for i in range(num_signals)]

    data = [np.zeros(total_samples[i]) for i in range(num_signals)]

    # Read all data records
    for record_idx in range(num_records):
        for signal_idx in range(num_signals):
            n_samples = samples_per_record[signal_idx]

            if bytes_per_sample == 2:
                raw_samples = struct.unpack(f'<{n_samples}h', f.read(n_samples * 2))
            else:
                raw_bytes = f.read(n_samples * 3)
                raw_samples = []
                for i in range(0, len(raw_bytes), 3):
                    value = int.from_bytes(raw_bytes[i:i+3], byteorder='little', signed=True)
                    raw_samples.append(value)
            
            # Convert to physical values
            start_idx = record_idx * n_samples
            end_idx = start_idx + n_samples
            data[signal_idx][start_idx:end_idx] = np.array(raw_samples) * scaling_factors[signal_idx] + offsets[signal_idx]

    channel_names = [x.replace('.', '') for x in signal_headers['label']]
    data, channel_names, eeg_indices = _filter_plus_annotations(data=data, channel_names=channel_names)
    units = [signal_headers['physical_dim'][i] for i in eeg_indices]
    data = _convert_to_microvolts(data, units)
    sample_rate = _determine_sample_rate(samples_per_record=samples_per_record, header=header,
                                         eeg_indices=eeg_indices, channel_names=channel_names)
    return np.array(data), channel_names, sample_rate



from pathlib import Path
import numpy as np
from breesy.load import load_recording
from breesy.features import get_mean_power, get_mean_power_per_channel, get_normalized_covariance_matrix, \
    get_spatial_filter_from_covariances, get_csp_features
from breesy.recording import split_by_window_duration

REPO_ROOT = Path(__file__).parent.parent


def test_mean_power_alpha_band():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    rec = load_recording(str(datapath))
    epochs = split_by_window_duration(rec, window_duration=5.0)
    epoch_data = np.array([epoch.data for epoch in epochs])
    n_epochs, n_channels, _ = epoch_data.shape

    # Act
    alpha_power = get_mean_power(epoch_data, freq_range='alpha', sample_rate=rec.sample_rate)

    # Assert
    assert alpha_power.ndim == 1
    assert alpha_power.shape[0] == len(epochs)
    assert np.all(alpha_power > 0)
    assert np.all(np.isfinite(alpha_power))


def test_mean_power_per_channel_alpha_band():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    rec = load_recording(str(datapath))
    epochs = split_by_window_duration(rec, window_duration=5.0)
    epoch_data = np.array([epoch.data for epoch in epochs])
    n_epochs, n_channels, _ = epoch_data.shape

    # Act
    alpha_power_per_channel = get_mean_power_per_channel(epoch_data, freq_range='alpha', sample_rate=rec.sample_rate)

    # Assert
    assert alpha_power_per_channel.ndim == 2
    assert alpha_power_per_channel.shape == (n_epochs, n_channels)
    assert np.all(alpha_power_per_channel > 0)
    assert np.all(np.isfinite(alpha_power_per_channel))
    assert not np.allclose(alpha_power_per_channel[:, 0], alpha_power_per_channel[:, 1])


def test_mean_power_custom_range():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    rec = load_recording(str(datapath))
    epochs = split_by_window_duration(rec, window_duration=5.0)
    epoch_data = np.array([epoch.data for epoch in epochs])

    # Act
    power = get_mean_power(epoch_data, freq_range=(10, 15), sample_rate=rec.sample_rate)

    # Assert
    assert power.ndim == 1
    assert power.shape[0] == len(epochs)
    assert np.all(power > 0)
    assert np.all(np.isfinite(power))


def test_frequency_bands():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    rec = load_recording(str(datapath))
    epochs = split_by_window_duration(rec, window_duration=5.0)
    epoch_data = np.array([epoch.data for epoch in epochs])
    bands = ['delta', 'theta', 'alpha', 'beta', 'gamma']

    # Act & Assert
    for band in bands:
        power = get_mean_power(epoch_data, freq_range=band, sample_rate=rec.sample_rate)

        assert power.ndim == 1
        assert power.shape[0] == len(epochs)
        assert np.all(power > 0)
        assert np.all(np.isfinite(power))


def test_csp_workflow():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    rec = load_recording(str(datapath))
    all_epochs = split_by_window_duration(rec, window_duration=2.0)

    # split into two classes
    n_total = len(all_epochs)
    class1_epochs = all_epochs[:n_total // 2]
    class2_epochs = all_epochs[n_total // 2:n_total]

    # Act
    cov1 = get_normalized_covariance_matrix(class1_epochs)
    cov2 = get_normalized_covariance_matrix(class2_epochs)

    spatial_filter = get_spatial_filter_from_covariances(cov1, cov2)

    n_components = 2
    features1 = get_csp_features(spatial_filter, class1_epochs, n_components=n_components)
    features2 = get_csp_features(spatial_filter, class2_epochs, n_components=n_components)

    # Assert
    n_channels = rec.number_of_channels
    assert cov1.shape == (n_channels, n_channels)
    assert cov2.shape == (n_channels, n_channels)
    assert spatial_filter.shape == (n_channels, n_channels)
    assert features1.shape == (len(class1_epochs), 2 * n_components)
    assert features2.shape == (len(class2_epochs), 2 * n_components)

    # finite (no NaN or inf)
    assert np.all(np.isfinite(cov1))
    assert np.all(np.isfinite(cov2))
    assert np.all(np.isfinite(spatial_filter))
    assert np.all(np.isfinite(features1))
    assert np.all(np.isfinite(features2))

    # symmetric
    assert np.allclose(cov1, cov1.T)
    assert np.allclose(cov2, cov2.T)


from pathlib import Path
import logging
import pyxdf

from .recording import Recording
from .RecordingMetadata import RecordingMetadata
from .errors import BreesyError, BreesyInternalError
from .load import _leave_only_eeg_channels
from .type_hints import enforce_type_hints


@enforce_type_hints
def load_xdf(filename: str) -> Recording:
    pyxdf.pyxdf.logger.setLevel(logging.ERROR)
    streams, _ = pyxdf.load_xdf(filename, verbose=None)
    stream_types = [x['info']['type'] for x in streams]
    eeg_streams_ix = [i for i, x in enumerate(stream_types) if x[0] == "EEG"]  # TODO: can there be more than one in a single stream? Check all [0]
    if len(eeg_streams_ix) > 1:
        raise BreesyError("Found more than one EEG stream in file.", "Currently, Breesy does not support multiple EEG streams in single .xdf file, sorry :(")
    eeg_stream = streams[eeg_streams_ix[0]]

    channel_count = int(eeg_stream['info']['channel_count'][0])
    sample_rate = int(eeg_stream['info']['nominal_srate'][0])
    channel_names = [row['label'][0] for row in eeg_stream['info']['desc'][0]['channels'][0]['channel']]
    if len(channel_names) != channel_count:
        raise BreesyInternalError("Number of channels does not correspond to number of extracted channel names.")
    data_array = eeg_stream['time_series'].T
    data_array, channel_names = _leave_only_eeg_channels(data_array, channel_names)

    metadata = RecordingMetadata(
        name=Path(filename).stem,
        file_path=Path(filename),
        description=f"XDF data loaded from {filename}"
    )

    return Recording(
        data=data_array,
        channel_names=channel_names,
        sample_rate=sample_rate,
        metadata=metadata
    )



from collections.abc import Callable
import numpy as np
import scipy
from tqdm.auto import tqdm

from .recording import Recording
from .breesy_scipy import _si_welch
from .constants import CLASSIC_BANDWIDTHS
from .errors import BreesyError, BreesyInternalError, protect_from_lib_error
from .type_hints import enforce_type_hints


@enforce_type_hints
def get_mean_power(epochs: np.ndarray, freq_range: str | tuple[int | float, int | float],
                   sample_rate: int | float) -> np.ndarray:
    """Extract mean power of a given frequency range from epoched EEG data.

    :param epochs: 3D array of epoched data (n_epochs, n_channels, n_samples)
    :param sample_rate: Sampling rate in Hz
    :param freq_range: Either a name of a classical band (e.g. "alpha"), or a frequency range (low, high) in Hz

    :return: 1D array of mean power values for each epoch
    """

    # TODO: instead of np.ndarray, epochs should be Recording type

    if isinstance(freq_range, str):
        freq_range = _get_freq_range_from_name(band_name=freq_range)

    mean_power = _get_power_feature(epochs=epochs, freq_range=freq_range, sample_rate=sample_rate, mode='mean')
    return mean_power


@enforce_type_hints
def get_mean_power_per_channel(epochs: np.ndarray,
                               freq_range: str | tuple[int | float, int | float],
                               sample_rate: int | float) -> np.ndarray:
    """Extract mean power per channel of a given frequency range from epoched EEG data.

    :param epochs: 3D array of epoched data (n_epochs, n_channels, n_samples)
    :param sample_rate: Sampling rate in Hz
    :param freq_range: Either a name of a classical band (e.g. "alpha"), or a frequency range (low, high) in Hz

    :return: 2D array of mean power values with shape (n_epochs, n_channels)
    """
    if isinstance(freq_range, str):
        freq_range = _get_freq_range_from_name(band_name=freq_range)

    mean_power_per_channel = _get_power_feature(
        epochs=epochs,
        freq_range=freq_range,
        sample_rate=sample_rate,
        mode='mean',
        per_channel=True
    )
    return mean_power_per_channel


@enforce_type_hints
def get_normalized_covariance_matrix(epochs: list[Recording]) -> np.ndarray:
    """Compute normalized covariance matrix from epoched EEG data.

    :param epochs: List of Recording objects representing one epoched Recording data

    :return: Normalized covariance matrix of shape (n_channels, n_channels)
    """
    if not epochs:
        raise BreesyError(
            "Cannot compute covariance from an empty list of epochs.",
            "Make sure that the epochs list is not empty. "
        )

    n_channels = epochs[0].number_of_channels
    for i, epoch in enumerate(epochs):
        if epoch.number_of_channels != n_channels:
            raise BreesyError(
                f"All epochs must have the same number of channels. "
                f"Epoch 0 has {n_channels} channels, but epoch {i} has {epoch.number_of_channels} channels.",
                "Ensure all epochs come from the same recording."
            )

    eeg_data = np.array([epoch.data for epoch in epochs])

    cov = []
    for i in range(eeg_data.shape[0]):
        epoch_cov = eeg_data[i] @ eeg_data[i].T
        cov.append(epoch_cov / np.trace(epoch_cov))

    return np.mean(np.array(cov), axis=0)


@enforce_type_hints
def get_csp_features(spatial_filter: np.ndarray, epochs: list[Recording], n_components: int = 2) -> np.ndarray:
    """Extract CSP features from epochs using a spatial filter.

    :param spatial_filter: Spatial filter matrix from get_spatial_filter_from_covariances()
    :param epochs: List of Recording objects representing epoched data
    :param n_components: Number of CSP components to use (default: 2, uses first and last n_components)

    :return: Feature matrix of shape (n_epochs, 2*n_components) containing log-variance features
    """
    if not epochs:
        raise BreesyError(
            "Cannot extract features from an empty list of epochs.",
            "Make sure that the epochs list is not empty."
        )

    if n_components < 1:
        raise BreesyError(
            f"n_components must be at least 1, got {n_components}.",
            "Set n_components to a positive integer (typically 2-4 for CSP)."
        )

    n_channels = epochs[0].number_of_channels
    if spatial_filter.shape[1] != n_channels:
        raise BreesyError(
            f"Spatial filter has {spatial_filter.shape[1]} columns but epochs have {n_channels} channels.",
            "Ensure the spatial filter was computed from epochs with the same number of channels."
        )

    # Stack epochs into 3D array
    eeg_data = np.array([epoch.data for epoch in epochs])

    return _get_Z_features(spatial_filter, eeg_data, n_components)

@enforce_type_hints
def get_spatial_filter_from_covariances(cov_class1: np.ndarray, cov_class2: np.ndarray) -> np.ndarray:
    """Compute spatial filter from two class covariance matrices using CSP.

    :param cov_class1: Covariance matrix for class 1 (e.g., left hand imagery)
    :param cov_class2: Covariance matrix for class 2 (e.g., right hand imagery)

    :return: Spatial filter matrix W of shape (n_channels, n_channels)
    """
    if cov_class1.shape != cov_class2.shape:
        raise BreesyError(
            f"Covariance matrices must have the same shape. "
            f"Got {cov_class1.shape} and {cov_class2.shape}.",
            "Ensure both classes have the same number of channels."
        )

    if cov_class1.ndim != 2 or cov_class1.shape[0] != cov_class1.shape[1]:
        raise BreesyError(
            f"Covariance matrices must be square 2D arrays. Got shape {cov_class1.shape}.",
            "Use get_normalized_covariance_matrix() to compute covariance matrices from epochs."
        )

    # decompose covariance matrices
    cov_comp = cov_class1 + cov_class2
    eigval, eigvec = _decompose_cov(cov_comp)

    # get whitening matrix
    P = _white_matrix(eigval, eigvec)

    # get and decompose S matrix
    s_left = _compute_S(cov_class1, P)
    s_left_eigvec, _ = _decompose_S(s_left, order='descending')

    # getting spatial filter W - we only need one of two!
    W = _spatial_filter(s_left_eigvec, P)

    return W


@protect_from_lib_error("numpy")
def _get_power_feature(epochs: np.ndarray, freq_range: tuple[int | float, int | float],
                       sample_rate: int | float, mode: str, per_channel: bool = False) -> np.ndarray:
    axis = None
    if per_channel:
        axis = -1

    # TODO: vectorize
    power_feature = []
    epoch_iter = tqdm(epochs, desc="Processing epochs", disable=len(epochs) < 10)

    for epoch in epoch_iter:
        spectral_piece = _get_spectral_epoch_piece(epoch=epoch, freq_range=freq_range, sample_rate=sample_rate)

        if mode == 'mean':
            result = np.mean(spectral_piece, axis=axis)
        elif mode == 'max':
            result = np.max(spectral_piece, axis=axis)
        else:
            raise BreesyInternalError(f'Aggregation mode "{mode}" is not supported.', ex=None)

        power_feature.append(result)

    return np.array(power_feature)

@protect_from_lib_error("numpy")
def _get_spectral_epoch_piece(epoch: np.ndarray, freq_range: tuple[int | float, int | float],
                              sample_rate: int | float):
    freqs, spectral = _si_welch(epoch, sample_rate=sample_rate)
    start_idx = np.searchsorted(freqs, freq_range[0])
    end_idx = np.searchsorted(freqs, freq_range[1])
    return spectral[:, start_idx:end_idx]


def _get_freq_range_from_name(band_name: str) -> tuple[int | float, int | float]:
    if band_name in CLASSIC_BANDWIDTHS.keys():
        return CLASSIC_BANDWIDTHS[band_name]
    else:
        supported_bands_str = ', '.join(f'"{x}"' for x in [CLASSIC_BANDWIDTHS.keys()])
        raise BreesyError(f'The provided name of the band - "{band_name}" - is unknown to Breesy.',
                            f"Use one of supported names: {supported_bands_str} "
                            'or provide frequency range as two numbers: (low, high).')


@protect_from_lib_error("numpy")
def _decompose_cov(avg_cov: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    lam, V = np.linalg.eig(avg_cov)
    lam_dsc = np.sort(lam)[::-1] # Sort eigenvalue descending order, default is ascending order sort
    idx_dsc = np.argsort(lam)[::-1] # Find index in descending order
    V_dsc = V[:, idx_dsc] # Sort eigenvectors descending order
    lam_dsc = np.diag(lam_dsc) # Diagonalize lam_dsc
    return lam_dsc, V_dsc


@protect_from_lib_error("scipy")
def _white_matrix(lam_dsc: np.ndarray, V_dsc: np.ndarray) -> np.ndarray:
    inv = scipy.linalg.inv(lam_dsc)
    lam_dsc_sqr = scipy.linalg.sqrtm(inv)
    return lam_dsc_sqr @ V_dsc.T


@protect_from_lib_error("numpy")
def _compute_S(avg_cov: np.ndarray, white: np.ndarray) -> np.ndarray:
    return white @ avg_cov @ white.T


@protect_from_lib_error("numpy")
def _decompose_S(S_one_class: np.ndarray, order: str = 'descending') -> tuple[np.ndarray, np.ndarray]:
    # Decompose S
    lam, B = np.linalg.eig(S_one_class)
    # Sort eigenvalues either descending or ascending
    if order == 'ascending':
        idx = lam.argsort()
    else:
        idx = lam.argsort()[::-1]
    return B[:, idx], lam[idx]


@protect_from_lib_error("numpy")
def _spatial_filter(B: np.ndarray, P: np.ndarray) -> np.ndarray:
    return B.T @ P

def _get_Z_features(spatial_filter, epochs_stacked, n_components: int) -> np.ndarray:
  Z = _compute_Z(spatial_filter, epochs_stacked, n_components)
  return _feat_vector(Z)


@protect_from_lib_error("numpy")
def _compute_Z(W: np.ndarray, E: np.ndarray, m: int) -> np.ndarray:
    Z = []
    W = np.delete(W, np.s_[m:-m:], 0)
    for i in range(E.shape[0]):
        Z.append(W @ E[i])
    return np.array(Z)


@protect_from_lib_error("numpy")
def _feat_vector(Z: np.ndarray) -> np.ndarray:
    feat = []
    for i in range(Z.shape[0]):
        var = np.var(Z[i], ddof=1, axis=1)
        varsum = np.sum(var)
        feat.append(np.log10(var / varsum))
    return np.array(feat)


from pathlib import Path
import numpy as np
from .RecordingMetadata import RecordingMetadata
from .errors import BreesyError
from .load import _calculate_sample_rate_from_time_column, _detect_time_column, _find_keys_in_data, _leave_only_eeg_channels
from .breesy_scipy import _loadmat
from .recording import Recording
from .log import logger
from .type_hints import enforce_type_hints


@enforce_type_hints
def load_mat(filename: str, sample_rate: int | float | None = None) -> Recording:
    """Load a MATLAB .mat file.

    :param filename: Path to the .mat file
    :param sample_rate: Sampling rate in Hz

    :return: Recording object containing the loaded data
    """
    mat_data = _loadmat(filename)

    meaningful_keys = _mat_extract_meaningful_keys(mat_data)

    data_key = None

    if len(meaningful_keys) == 1:
        data_key = meaningful_keys[0]
    else:
        # Find the largest array that looks like EEG data
        largest_size = 0
        for key in meaningful_keys:
            value = mat_data[key]
            if isinstance(value, np.ndarray) and value.size > largest_size and value.ndim >= 2:
                largest_size = value.size
                data_key = key

    data = mat_data[data_key]  # TODO: there can be 'O' types, as AI previously suggested?
    if not isinstance(data, np.ndarray):
        raise BreesyError(
            f"Problem reading data which is of type {type(data)}. It may be that Breesy failed to read the file correctly."  # TODO: test different files
        )
    if data.ndim != 2:
        raise BreesyError(
            "Found data has an incorrect number of dimensions. "
            "Expected data.ndim to be 2, but got "
            f"{data.ndim} dimensions instead.",
            "The data may be not EEG data, "
            "there are several records joined into one file, "
            "or Breesy did not found the correct data array in the file."
        )

    if data.shape[0] > data.shape[1]:
        logger.warning("The first data dimension seems to be the samples, because data contains more rows than columns. Dimensions will be switched places.")
        data = data.T
    n_channels, n_samples = data.shape
    default_channel_names = [f'Channel_{i + 1}' for i in range(data.shape[0])]

    # find or create channel names
    channel_name_keys = _find_keys_in_data(meaningful_keys, ['ch', 'ch name', 'ch names', 'channel names', 'channels', 'names', '10-20', '10-10', 'pos', 'positions'])
    if len(channel_name_keys) == 0:
        channel_names = default_channel_names
        channel_name_key = None
    elif len(channel_name_keys) > 1:
        logger.warning(
            "found several keys in data which may contain channel names: "
            f"{channel_name_keys}. "
            "Using the first one.")
        channel_name_key = channel_name_keys[0]
    else:
        channel_name_key = channel_name_keys[0]

    if channel_name_key:
        found_names = mat_data[channel_name_key]
        if not isinstance(found_names, (list, tuple, np.ndarray)):
            logger.warning(
                "found channel name information does not appear to be a list: "
                f"{found_names}. "
                "Default channel names will be used instead.")
            channel_names = default_channel_names
        elif (n_channels != len(found_names)) and (n_samples != len(found_names)):
            logger.warning(
                f"the number of channel names ({len(found_names)}) provided does not match the "
                f"number of channels ({n_channels}) in the EEG data. "
                "Default channel names will be used instead."
            )
            channel_names = default_channel_names
        elif (n_channels != len(channel_names)) and (n_samples == len(channel_names)):
            logger.warning('data rows and columns will be switched places according to the number of found channel names.')
            data = data.T
            n_channels, n_samples = data.shape
            channel_names = found_names
        else:
            channel_names = found_names

    # try to find it in keys
    extracted_sample_rate = _detect_sample_rate_in_mat_keys(mat_data, meaningful_keys)
    if extracted_sample_rate is None:
        # else in time column
        time_column_idx = _detect_time_column(data)
        if time_column_idx is not None:
            extracted_sample_rate = _calculate_sample_rate_from_time_column(data[time_column_idx])
            logger.info(f"Detected time column at index {time_column_idx}. Extracted sample rate: {extracted_sample_rate} Hz")

    # Determine final sample rate
    if sample_rate is not None:
        if extracted_sample_rate is not None and abs(sample_rate - extracted_sample_rate) > 0.01:
            logger.warning(f"Warning: Provided sample_rate ({sample_rate} Hz) differs from extracted sample rate ({extracted_sample_rate} Hz). Using the provided sample_rate.")
        sample_rate = float(sample_rate)
    elif extracted_sample_rate is not None:
        sample_rate = extracted_sample_rate
    else:
        raise BreesyError(
            "No sample rate provided and could not detect sample rate from data.",
            "Please provide the sample rate manually by passing the 'sample_rate' parameter. "
            "Example: load_mat(filename, sample_rate=250)"
        )

    data, channel_names = _leave_only_eeg_channels(data, channel_names)
    # channel names may not start from 1 after this, but maybe we should leave it like this

    metadata = RecordingMetadata(
        name=Path(filename).stem,
        file_path=Path(filename),
        description=f"MATLAB data loaded from {filename}"
    )

    return Recording(
        data=data,
        channel_names=channel_names,
        sample_rate=sample_rate,
        events=None,  # No events for now - could be added as a separate function
        metadata=metadata
    )


def _mat_extract_meaningful_keys(mat_data: dict) -> list[str]:
    keys = [k for k in mat_data.keys() if not k.startswith('__')]
    if not keys:
        raise BreesyError(
            "No valid data found. The MATLAB file appears to be empty or contains only metadata."
        )
    elif len(keys) == 1:
        logger.warning(
            "data file likely contains only the recording and "
            "no useful metadata. Please find the sample rate and channel names "
            "in your data description."
        )  # TODO: fast instructions or link to documentation
    return keys


def _detect_sample_rate_in_mat_keys(mat_data: dict, keys: list[str]) -> float | None:
    sample_rate_keys = _find_keys_in_data(keys, ['sample rate', 'sp', 'sr', 'hz', 'sampling rate', 'sfreq', 'fs', 'srate'])
    if len(sample_rate_keys) == 0:
        return None

    sample_rate_key = sample_rate_keys[0]
    if len(sample_rate_keys) > 1:
        logger.warning(
            "found several keys in data which may contain sample rate:"
            f"\n{sample_rate_keys}\n"
            "Using the first one.")

    sr_value = mat_data[sample_rate_key]
    try:
        if isinstance(sr_value, (int, float, str)):
            return float(sr_value)
        elif isinstance(sr_value, np.ndarray) and sr_value.size == 1:
            return float(sr_value.item())
        elif isinstance(sr_value, (list, tuple)) and len(sr_value) == 1:
            return float(sr_value[0])
        else:
            raise ValueError
    except ValueError:
        logger.warning(
            f"Information found in column '{sample_rate_key}' "
            f"could not be read automatically; column contents are:\n{sr_value}"
        )
        return None